/*
 * Show lines from a file.
 *
 * Copyright (C) 2022, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* NOTE: For inclusion in the C library: stream acquisition and access. */

#include <fsclient/client.h>
#include <systypes/env.h>
#include <systypes/fcntl.h>

static file_t *input, *output;
static char output_buffer[256];

#define _fprintf(file, s, ...) \
( \
  sprintf(output_buffer, s, ##__VA_ARGS__), \
  client_write(file, output_buffer, strlen(output_buffer)) \
)



/* Read a line from the file. */

static int readline(file_t *file, int *lineno, char **start, char **end)
{
  static char buf[256];
  static offset_t current = 0, limit = 0;
  static int next_lineno = 1;
  char *newline;

  /* Advance to the next line if previously found. */

  if (next_lineno > *lineno)
    *lineno = next_lineno;

  /* Obtain file content. */

  if (!limit)
  {
    limit = client_read(file, buf, 256);

    if (!limit)
    {
      *start = NULL;
      *end = NULL;
      return 0;
    }
  }

  /* Find newline. */

  newline = (char *) memchr(buf + current, (int) '\n', limit - current);
  *start = (buf + current);

  /* Return final line fragment, potentially leaving more available text. */

  if (newline != NULL)
  {
    /* Advance the line number when reading again. */

    next_lineno = (*lineno) + 1;

    /* Define the end of the line and the start of the next line. */

    *end = newline + 1;
    current = *end - buf;

    if (current >= limit)
    {
      current = 0;
      limit = 0;
    }

    return 1;
  }

  /* Or return mid-line fragment reaching the limit of the available text. */

  else
  {
    *end = buf + limit;
    current = 0;
    limit = 0;

    return 1;
  }
}



int main(int argc, char *argv[])
{
  file_t *file;
  int lineno, startline, numlines;
  char *start, *end;

  /* NOTE: For inclusion in the C library: stream acquisition and access. */

  input = client_get_stream(ENV_INPUT_STREAM_NAME, O_RDONLY);
  output = client_get_stream(ENV_OUTPUT_STREAM_NAME, O_WRONLY);

  if (argc < 4)
    return 1;

  if (!strcmp(argv[1], "-"))
    file = input;
  else
    file = client_open(argv[1], O_RDONLY);

  if (!client_opened(file))
  {
    if (file != NULL)
      _fprintf(output, "Error: %s\n", l4sys_errtostr(file->error));
    else
      _fprintf(output, "Could not open file.\n");

    client_flush(output);
    return 1;
  }

  startline = atoi(argv[2]);
  numlines = atoi(argv[3]);

  lineno = 1;

  while (lineno < startline + numlines)
  {
    if (!readline(file, &lineno, &start, &end))
    {
      _fprintf(output, "EOF error at line %d.\n", lineno);
      client_flush(output);
      return 1;
    }

    if ((lineno >= startline) && (lineno < startline + numlines))
      client_write(output, start, end - start);
  }

  client_close(file);

  /* NOTE: For inclusion in the C library: stream acquisition and access. */

  client_flush(output);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
